from .base import Widget


class NumberInput(Widget):
    '''
    Widget for input of a number
    '''
    def __init__(self, id=None, style=None, min_value=0, max_value=100, step=1, **ex):
        '''
        Instantiate a new instance of the Number input widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`
        
        :param min_value:   Minimum value (default 0)
        :type min_value:    ``int``

        :param max_value:   Maximum value (default 100)
        :type max_value:    ``int``

        :param step:        Step of the adjustment buttons
        :type step:         ``int``
        '''
        self._min_value = min_value
        self._max_value = max_value
        self._step = step
        super().__init__(id=id, style=style, min_value=min_value, max_value=max_value, step=step, **ex)

    def _configure(self, **kw):
        pass

    @property
    def value(self):
        '''
        Current value
        
        :rtype: ``int``
        :return: The current value
        '''
        return self._get_value()

    @value.setter
    def value(self, value):
        self._set_value(value)



def info(window, title, message):
    raise NotImplementedError()


def question(window, title, message):
    raise NotImplementedError()


def confirm(window, title, message):
    raise NotImplementedError()


def error(window, title, message):
    raise NotImplementedError()


def stack_trace(window, title, message, content, retry=False):
    raise NotImplementedError()


def save_file(window, title, suggested_filename, file_types):
    raise NotImplementedError()



from .base import Widget
from toga.constants import *


class Label(Widget):
    '''
    Label widget
    '''
    def __init__(self, text, id=None, style=None, alignment=LEFT_ALIGNED):
        '''
        Instantiate a new instance of the label widget

        :param text:        Text of the label
        :type  text:        ``str``

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param alignment:   Alignment of the label, default is left. Alignments can be found
                            in toga.constants
        :type alignment:    ``int``
        '''
        super().__init__(id=id, style=style, text=text, alignment=alignment)

    def _configure(self, text, alignment):
        self.text = text
        self.alignment = alignment

    @property
    def alignment(self):
        '''
        The alignment of the label text

        :returns: Alignment of the label, default is left. Alignments can be found
                            in toga.constants
        :rtype: ``int``
        '''
        return self._alignment

    @alignment.setter
    def alignment(self, value):
        self._alignment = value
        self._set_alignment(value)

    @property
    def text(self):
        '''
        The text of the label

        :rtype: ``str``
        '''
        return self._text

    @text.setter
    def text(self, value):
        if value is None:
            self._text = ''
        else:
            self._text = str(value)
        self._set_text(value)
        self.rehint()

    def _set_alignment(self, value):
        raise NotImplementedError('Label widget must define _set_alignment()')

    def _set_text(self, value):
        raise NotImplementedError('Label widget must define _set_text()')



from .app import App
from .window import Window
from .font import Font

# Widgets
from .widgets.box import *
from .widgets.button import *
from .widgets.detailedlist import *
from .widgets.icon import *
# from .widgets.image import *
from .widgets.imageview import *
from .widgets.label import *
# from .widgets.dialog import *
from .widgets.multilinetextinput import *
from .widgets.navigationview import *
from .widgets.optioncontainer import *
# from .widgets.passwordinput import *
from .widgets.progressbar import *
from .widgets.scrollcontainer import *
from .widgets.splitcontainer import *
from .widgets.table import *
from .widgets.textinput import *
from .widgets.tree import *
from .widgets.webview import *
from .widgets.selection import Selection
from .widgets.numberinput import NumberInput

__all__ = [
    'App',
    'Window',

    # Fonts
    'Font',

    # WIDGETS
    'Box',
    'Button',
    'DetailedList',
    'Icon',
    # 'Image',
    'ImageView',
    'Label',
    # 'Dialog',
    'MultilineTextInput',
    'NavigationView',
    'OptionContainer',
    # 'PasswordInput',
    'ProgressBar',
    'ScrollContainer',
    'Selection',
    'SplitContainer',
    'NumberInput',
    'Table',
    'TextInput',
    'Tree',
    'WebView',
]



from .base import Widget


class ScrollContainer(Widget):
    '''
    A scrollable container
    '''
    _CONTAINER_CLASS = None

    def __init__(self, id=None, style=None, horizontal=True,
                 vertical=True, content=None):
        '''
        Instantiate a new instance of the scrollable container widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param horizontal: Enable horizontal scroll bar
        :type  horizontal: ``bool``

        :param vertical: Enable vertical scroll bar
        :type  vertical: ``bool``

        :param content: The content of the scroll window
        :type  content: :class:`toga.Widget`
        '''
        super().__init__(id=id, style=style, horizontal=horizontal, vertical=vertical, content=content)

    def _configure(self, horizontal, vertical, content):
        self.horizontal = horizontal
        self.vertical = vertical
        self.content = content

    @property
    def content(self):
        '''
        Content of the scroll window

        :rtype: :class:`toga.Widget`
        '''
        return self._content

    @content.setter
    def content(self, widget):
        self._content = widget
        if widget:
            widget._update_layout()
            self._content.app = self.app
            self._content.window = self.window

            if widget._impl is None:
                self._inner_container = self._CONTAINER_CLASS()
                self._inner_container.root_content = widget
            else:
                self._inner_container = widget

            self._set_content(self._inner_container, widget)

    @property
    def vertical(self):
        '''
        Enable vertical scaling

        :rtype: ``bool``
        '''
        return self._vertical

    @vertical.setter
    def vertical(self, value):
        self._vertical = value
        self._set_vertical(value)

    @property
    def horizontal(self):
        '''
        Enable horizontal scaling

        :rtype: ``bool``
        '''
        return self._horizontal

    @horizontal.setter
    def horizontal(self, value):
        self._horizontal = value
        self._set_horizontal(value)

    def _set_vertical(self, value):
        raise NotImplementedError('ScrollContainer must define _set_vertical()')

    def _set_horizontal(self, value):
        raise NotImplementedError('ScrollContainer must define _set_horizontal()')



from .base import Widget


class TextInput(Widget):
    '''
    Text input widget
    '''
    def __init__(self, id=None, style=None, initial=None, placeholder=None, readonly=False):
        '''
        Instantiate a new instance of the text input widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param initial: The initial text
        :type  initial: ``str``
        
        :param placeholder: The placeholder text
        :type  placeholder: ``str``
        
        :param readonly: Whether a user can write into the text input, defaults to `False`
        :type  readonly: ``bool``
        '''
        super().__init__(id=id, style=style, initial=initial, placeholder=placeholder, readonly=readonly)

    def _configure(self, initial, placeholder, readonly):
        self.readonly = readonly
        self.placeholder = placeholder
        self.value = initial
        self.rehint()

    @property
    def readonly(self):
        '''
        Whether a user can write into the text input
        
        :rtype: ``bool``
        '''
        return self._readonly

    @readonly.setter
    def readonly(self, value):
        self._readonly = value
        self._set_readonly(value)

    @property
    def placeholder(self):
        '''
        The placeholder text
        
        :rtype: ``str``
        '''
        return self._placeholder

    @placeholder.setter
    def placeholder(self, value):
        if value is None:
            self._placeholder = ''
        else:
            self._placeholder = str(value)
        self._set_placeholder(value)

    @property
    def value(self):
        '''
        The value of the text input field
        
        :rtype: ``str``
        '''
        return self._get_value()

    @value.setter
    def value(self, value):
        if value is None:
            v = ''
        else:
            v = str(value)
        self._set_value(v)
        self.rehint()

    def clear(self):
        '''
        Clear the value
        '''
        self.value = ''

    def _set_readonly(self, value):
        raise NotImplementedError('TextInput widget must define _set_readonly()')

    def _set_placeholder(self, value):
        raise NotImplementedError('TextInput widget must define _set_placeholder()')

    def _set_value(self, value):
        raise NotImplementedError('TextInput widget must define _set_value()')



from .base import Widget


class DetailedList(Widget):
    def __init__(self, id=None, data=None, on_delete=None, on_refresh=None, style=None):
        super().__init__(id=id, data=data, on_delete=on_delete, on_refresh=on_refresh, style=style)

    def _configure(self, data, on_delete, on_refresh):
        self.data = data
        self.on_delete = on_delete
        self.on_refresh = on_refresh

    def add(self, item):
        self.data.append(item)
        self._add(item)



from .base import Widget


class WebView(Widget):
    '''
    Web view widget
    '''
    def __init__(self, id=None, style=None, url=None, on_key_down=None):
        '''
        Instantiate a new instance of the tree widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`
        
        :param url: The URL to start with
        :type  url: ``str``
        
        :param on_key_down: The callback method for when the a key is pressed within
            the web view
        :type  on_key_down: ``callable``
        '''
        super(WebView, self).__init__(id=id, style=style, url=url, on_key_down=on_key_down)

    def _configure(self, url, on_key_down):
        self.url = url
        self.on_key_down = on_key_down

    @property
    def url(self):
        '''
        The current URL
        
        :rtype: ``str``
        '''
        return self._url

    @url.setter
    def url(self, value):
        self._url = value
        self._set_url(value)

    def set_content(self, root_url, content):
        '''
        Set the content of the web view
        
        :param root_url: The URL
        :type  root_url: ``str``
        
        :param content: The new content
        :type  content: ``str``
        '''
        self._url = root_url
        self._set_content(root_url, content)

    def _set_url(self, value):
        raise NotImplementedError('Webview widget must define _set_url()')

    def _set_content(self, root_url, content):
        raise NotImplementedError('Webview widget must define _set_content()')

    def evaluate(self, javascript):
        '''
        Evaluate a JavaScript expression
        
        :param javascript: The javascript expression
        :type  javascript: ``str``
        '''
        raise NotImplementedError('Webview widget must define evaluate()')



import importlib
import os
import sys
import types

from .constants import *

# Work around import loop issues (toga -> platform -> toga.interface) import
# all these things before we import the platform stuff
import toga.interface.app  # NOQA

__all__ = [
    '__version__',
    'platform'
]

# Examples of valid version strings
# __version__ = '1.2.3.dev1'  # Development release 1
# __version__ = '1.2.3a1'     # Alpha Release 1
# __version__ = '1.2.3b1'     # Beta Release 1
# __version__ = '1.2.3rc1'    # RC Release 1
# __version__ = '1.2.3'       # Final Release
# __version__ = '1.2.3.post1' # Post Release 1

__version__ = '0.2.12'

platform = None


def get_platform_name():
    """
    Get the name of the current platform
    :return: The name of the platform, e.g. tvOS
    :rtype: ``str``
    """
    platform_name = os.environ.get('TOGA_PLATFORM')

    # If we don't have a manually defined platform, attempt to
    # autodetect and set the platform
    if platform_name is None:
        if sys.platform == 'ios':
            platform_name = 'iOS'
        elif sys.platform == 'tvos':
            platform_name = 'tvOS'
        elif sys.platform == 'watchos':
            platform_name = 'watchOS'
        elif sys.platform == 'android':
            platform_name = 'android'
        elif sys.platform == 'darwin':
            platform_name = 'cocoa'
        elif sys.platform == 'linux':
            platform_name = 'gtk'
        elif sys.platform == 'win32':
            platform_name = 'winforms'
        else:
            raise RuntimeError("Couldn't identify a supported host platform.")
    return platform_name


def set_platform(module_name=None, local_vars=locals()):
    "Configures toga to use the specified platform module"
    # Note - locals is deliberately passed in as an argument; because it is
    # a dictionary, this results in the module level locals dictionary being
    # bound as a local variable in this method -- and a persistent one,
    # because it was evaluated and bound at time of method import.

    # First check for an environment variable setting the platform
    if module_name is None:
        platform_name = get_platform_name()
        module_name = 'toga_' + platform_name

    # # Purge any existing platform symbols in the toga module
    # if local_vars['platform']:
    #     for symbol in local_vars['platform'].__all__:
    #         # Exclude __version__ from the list of symbols that is
    #         # ported, because toga itself has a __version__ identifier.
    #         if symbol != '__version__':
    #             # Remove any modules from the importable module list
    #             if isinstance(local_vars[symbol], types.ModuleType):
    #                 del sys.modules['toga.%s' % symbol]
    #             local_vars.pop(symbol)

    # Import the new platform module
    try:
        # Set the new platform module into the module namespace
        local_vars['platform'] = importlib.import_module(module_name)

        # Export all the symbols *except* for __version__ from the platform module
        # The platform has it's own version identifier.
        for symbol in local_vars['platform'].__all__:
            if symbol == '__version__':
                if local_vars['platform'].__version__ != __version__:
                    raise RuntimeError('Toga core is version %s; %s platform backend is version %s.' % (
                            __version__,
                            module_name,
                            local_vars['platform'].__version__
                        )
                    )
            else:
                local_vars[symbol] = getattr(local_vars['platform'], symbol)
                # Make sure any modules are added to the importable module list
                if isinstance(local_vars[symbol], types.ModuleType):
                    sys.modules['toga.%s' % symbol] = local_vars[symbol]
    except ImportError as e:
        if e.name == module_name:
            local_vars['platform'] = None
            print("Couldn't import %s platform module; try running 'pip install %s'." % (module_name[5:], module_name))
            sys.exit(-1)
        else:
            raise

# On first import, do an auto-detection of platform.
if get_platform_name() != 'dummy':
    set_platform()



class Window:
    '''
    Window
    '''
    _CONTAINER_CLASS = None
    _DIALOG_MODULE = None

    def __init__(self, id_=None, title=None,
                 position=(100, 100), size=(640, 480),
                 toolbar=None, resizeable=True,
                 closeable=True, minimizable=True):
        '''
        Instantiates a window

        :param id_: The ID of the window (optional)
        :type  id_: ``str``

        :param title: Title for the window (optional)
        :type  title: ``str``

        :param position: Position of the window, as x,y coordinates
        :type  position: ``tuple`` of (``int``, ``int``)

        :param size: Size of the window, as (width, height) sizes, in pixels
        :type  size: ``tuple`` of (``int``, ``int``)

        :param toolbar: An list of widgets to add to a toolbar
        :type  toolbar: ``list`` of :class:`toga.Widget`

        :param resizable: Toggle if the window is resizable by the user, defaults
            to `True`.
        :type  resizable: ``bool``

        :param closable: Toggle if the window is closable by the user, defaults
            to `True`.
        :type  closable: ``bool``

        :param minimizable: Toggle if the window is minimizable by the user, defaults
            to `True`.
        :type  minimizable: ``bool``
        '''
        if self._CONTAINER_CLASS is None:
            raise NotImplementedError('Window class must define show()')

        self._id = id_ if id_ else id(self)
        self._impl = None
        self._app = None
        self._container = None
        self._content = None

        self.position = position
        self.size = size

        self.resizeable = resizeable
        self.closeable = closeable
        self.minimizable = minimizable

        self._config = {
            'title': title,
            'position': position,
            'size': size,
            'toolbar': toolbar,
            'resizeable': resizeable,
            'closeable': closeable,
            'minimizable': minimizable,
        }

    def _create(self):
        self.create()
        self._configure(**self._config)

    def _configure(self, title, position, size, toolbar, resizeable, closeable, minimizable):
        self.title = title
        self.toolbar = toolbar

    @property
    def app(self):
        '''
        Instance of the :class:`toga.App` that this window belongs to

        :rtype: :class:`toga.App`
        '''
        return self._app

    @app.setter
    def app(self, app):
        if self._app:
            raise Exception("Window is already associated with an App")

        self._app = app
        self._set_app(app)

    def _set_app(self, app):
        pass

    @property
    def title(self):
        '''
        Title of the window

        :rtype: ``str``
        '''
        return self._title

    @title.setter
    def title(self, title):
        if not title:
            title = "Toga"

        self._set_title(title)
        self._title = title

    @property
    def toolbar(self):
        '''
        Toolbar for the window

        :rtype: ``list`` of :class:`toga.Widget`
        '''
        return self._toolbar

    @toolbar.setter
    def toolbar(self, items):
        # If there are toolbar items defined, add a toolbar to the window
        self._toolbar = items
        if self._toolbar:
            self._set_toolbar(items)

    def _set_toolbar(self, items):
        pass

    @property
    def content(self):
        '''
        Content of the window

        :rtype: :class:`toga.Widget`
        '''
        return self._content

    @content.setter
    def content(self, widget):
        # Save the content widget.
        widget._update_layout()

        # Assign the widget to the same app as the window.
        widget.app = self.app

        # Assign the widget to window.
        widget.window = self

        if widget._impl is None:
            self._container = self._CONTAINER_CLASS()
            self._container.content = widget
        else:
            self._container = widget

        self._content = widget

        self._set_content(widget)

    def _set_content(self, widget):
        pass

    @property
    def size(self):
        '''
        Size of the window, as width, height

        :rtype: ``tuple`` of (``int``, ``int``)
        '''
        return self._size

    @size.setter
    def size(self, size):
        self._size = size
        self._set_size(size)

    def _set_size(self, size):
        pass

    @property
    def position(self):
        '''
        Position of the window, as x, y

        :rtype: ``tuple`` of (``int``, ``int``)
        '''
        return self._position

    @position.setter
    def position(self, position):
        self._position = position
        self._set_position(position)

    def _set_position(self, position):
        pass

    def show(self):
        '''
        Show window, if hidden
        '''
        raise NotImplementedError('Window class must define show()')

    def on_close(self):
        pass

    def info_dialog(self, title, message):
        return self._DIALOG_MODULE.info(self, title, message)

    def question_dialog(self, title, message):
        return self._DIALOG_MODULE.question(self, title, message)

    def confirm_dialog(self, title, message):
        return self._DIALOG_MODULE.confirm(self, title, message)

    def error_dialog(self, title, message):
        return self._DIALOG_MODULE.error(self, title, message)

    def stack_trace_dialog(self, title, message):
        return self._DIALOG_MODULE.stack_trace(self, title, message)

    def save_file_dialog(self, title, suggested_filename, file_types):
        return self._DIALOG_MODULE.save_file(self, title, suggested_filename,
                                             file_types)



from .base import Widget


class SplitContainer(Widget):
    '''
    Split container widget
    '''
    _CONTAINER_CLASS = None
    HORIZONTAL = False
    VERTICAL = True

    def __init__(self, id=None, style=None, direction=VERTICAL, content=None):
        '''
        Instantiate a new instance of the split container widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param direction: The direction for the container split, either `SplitContainer.HORIZONTAL` or
            `SplitContainer.VERTICAL`
        :type  direction: ``bool``

        :param content: The list of components to be split
        :type  content: ``list`` of :class:`toga.Widget`
        '''
        super().__init__(id=id, style=style, direction=direction, content=content)
        self._direction = direction
        self._containers = []

    def _configure(self, direction, content):
        self.content = content
        self.direction = direction

    @property
    def content(self):
        '''
        The content of the split container

        :rtype: ``list`` of :class:`toga.Widget`
        '''
        return self._content

    @content.setter
    def content(self, content):
        if content is None:
            self._content = None
            return

        if len(content) < 2:
            raise ValueError('SplitContainer content must have at least 2 elements')

        self._content = content

        for position, widget in enumerate(self._content):
            widget._update_layout()
            widget.app = self.app
            widget.window = self.window

            if widget._impl is None:
                container = self._CONTAINER_CLASS()
                container.root_content = widget
                widget._container = container
            else:
                container = widget

            self._add_content(position, container)

            self._containers.append(container)

    def _set_app(self, app):
        if self._content:
            for content in self._content:
                content.app = self.app

    def _set_window(self, window):
        if self._content:
            for content in self._content:
                content.window = self.window

    @property
    def direction(self):
        '''
        The direction of the split

        :rtype: ``bool``
        '''
        return self._direction

    @direction.setter
    def direction(self, value):
        self._direction = value
        self._set_direction(value)
        self.rehint()

    def on_resize(self):
        pass



from .base import Widget


class Selection(Widget):
    '''
    Selection widget
    '''
    def __init__(self, id=None, style=None, items=list()):
        '''
        Instantiate a new instance of the selection widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param items:       Items for the selection:
        :type  items:       ``list`` of ``str``
        '''
        self._items = items
        super().__init__(id=id, style=style, items=items)

    def _configure(self, items):
        pass

    def _create(self):
        super()._create()
        for item in self._items:
            self._add_item(item)

    @property
    def items(self):
        '''
        The list of items
        
        :rtype: ``list`` of ``str``
        '''
        return self._items

    @items.setter
    def items(self, items):
        self._remove_all_items()

        for i in items:
            self._add_item(i)

        self._items = items

    @property
    def value(self):
        '''
        The value of the selected item
        
        :rtype: ``str``
        '''
        return self._get_selected_item()

    @value.setter
    def value(self, value):
        if value not in self._items:
            raise ValueError("Not an item in the list.")

        self._select_item(value)



from .base import Widget


class Button(Widget):
    """
    Button widget, a clickable button

    :param label:       Text to be shown on the button
    :type label:        ``str``

    :param id:          An identifier for this widget.
    :type  id:          ``str``

    :param style:       an optional style object. If no style is provided then a
                        new one will be created for the widget.
    :type style:        :class:`colosseum.CSSNode`

    :param on_press:    Function to execute when pressed
    :type on_press:     ``callable``
    """
    def __init__(self, label, id=None, style=None, on_press=None, enabled=None):
        super().__init__(id=id, style=style, label=label, on_press=on_press,
                        enabled=enabled)

    def _configure(self, label, on_press, enabled):
        self.label = label
        self.on_press = on_press
        self.enabled = enabled

    @property
    def label(self):
        """
        :returns: The label value
        :rtype: ``str``
        """
        return self._label

    @label.setter
    def label(self, value):
        """
        Set the label value

        :param value: The new label value
        :type  value: ``str``
        """
        if value is None:
            self._label = ''
        else:
            self._label = str(value)
        self._set_label(str(value))
        self.rehint()

    def _set_label(self, value):
        raise NotImplementedError('Button widget must define _set_label()')

    @property
    def on_press(self):
        """
        The callable function for when the button is pressed

        :rtype: ``callable``
        """
        return self._on_press

    @on_press.setter
    def on_press(self, handler):
        """
        Set the function to be executed on button press.

        :param handler:     callback function
        :type handler:      ``callable``
        """
        self._on_press = handler
        self._set_on_press(handler)

    def _set_on_press(self, value):
        pass

    @property
    def enabled(self):
        """
        Indicates whether the button can be pressed by the user.

        :returns:   Button status. Default is True.
        :rtype:     ``Bool`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        """
        Set if the button can be pressed by the user.

        :param value:   Enabled state for button
        :type value:    ``Bool``
        """
        if value is None:
            self._enabled = True
        else:
            self._enabled = value
        self._set_enabled(value)

    def _set_enabled(self, value):
        raise NotImplementedError('Button widget must define _set_enabled()')


from .base import Widget


class Table(Widget):
    '''
    Table widget
    '''
    def __init__(self, headings, id=None, style=None):
        '''
        Instantiate a new instance of the split container widget

        :param headings: The list of headings for the table
        :type  headings: ``list`` of ``str``

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`
        '''
        super().__init__(id=id, style=style)
        self.headings = headings

    def _configure(self):
        pass

    def insert(self, index, *data):
        '''
        Insert a new row into the table
        
        :param index: The index to insert at, the end if `None`
        :type  index: ``int`` or ``NoneType``
        
        :param *data: A list of values for each of the columns
        :type  *data: ``list`` of ``object``
        '''
        if len(data) != len(self.headings):
            raise Exception('Data size does not match number of headings')

        if index is None:
            self._data.append(data)
        else:
            self._data.insert(index, data)



from builtins import id as identifier
from colosseum import CSS


class Point:
    def __init__(self, top, left):
        self.top = top
        self.left = left

    def __repr__(self):
        return '<Point (%s,%s)>' % (self.left, self.top)


class Layout:
    def __init__(self, node, width=None, height=None, top=0, left=0):
        self.node = node
        self.width = width
        self.height = height
        self.top = top
        self.left = left
        self._dirty = True

    def __repr__(self):
        if self.node:
            return '<Layout%s (%sx%s @ %s,%s)>' % (
                {
                    True: ' (dirty)',
                    False: '',
                    None: ' (evaluating)'
                }[self._dirty],
                self.width, self.height,
                self.absolute.left, self.absolute.top
            )
        else:
            return '<Layout%s (%sx%s @ %s,%s)>' % (
                {
                    True: ' (dirty)',
                    False: '',
                    None: ' (evaluating)'
                }[self._dirty],
                self.width, self.height,
                self.left, self.top
            )

    def __eq__(self, value):
        return all([
            self.width == value.width,
            self.height == value.height,
            self.top == value.top,
            self.left == value.left
        ])

    def reset(self):
        self.width = None
        self.height = None
        self.top = 0
        self.left = 0

    ######################################################################
    # Layout dirtiness tracking.
    #
    # If dirty == True, the layout is known to be invalid.
    # If dirty == False, the layout is known to be good.
    # If dirty is None, the layout is currently being re-evaluated.
    ######################################################################
    @property
    def dirty(self):
        return self._dirty

    @dirty.setter
    def dirty(self, value):
        self._dirty = value
        for child in self.node.children:
            child.layout.dirty = value

    ######################################################################
    # Implied geometry properties
    ######################################################################
    @property
    def right(self):
        return self.left + self.width

    @property
    def bottom(self):
        return self.top + self.height

    @property
    def absolute(self):
        if self.node.parent:
            parent_layout = self.node.parent.layout
            return Point(
                top=parent_layout.origin.top + parent_layout.top + self.top,
                left=parent_layout.origin.left + parent_layout.left + self.left,
            )
        else:
            return Point(top=self.top, left=self.left)

    @property
    def origin(self):
        if self.node.parent:
            parent_layout = self.node.parent.layout
            return Point(
                top=parent_layout.origin.top + parent_layout.top,
                left=parent_layout.origin.left + parent_layout.left,
            )
        else:
            return Point(top=0, left=0)


class Widget:
    '''
    This is the base widget implementation that all widgets in Toga
    derive from.

    It defines the interface for core functionality for children, styling,
    layout and ownership by specific App and Window.

    Apart from the above, this is an abstract implementation which must
    be made concrete by some platform-specific code for the _apply_layout
    method.

    :param id:      An identifier for this widget.
    :type  id:      ``str``

    :param style:   An optional style object. If no style is provided then a
                    new one will be created for the widget.
    :type style:    :class:`colosseum.CSSNode`
    '''
    def __init__(self, id=None, style=None, **config):
        self._id = id if id else identifier(self)
        self._parent = None
        self._children = None
        self._window = None
        self._app = None
        self._impl = None
        self.__container = None
        self._layout_in_progress = False

        self._config = config

        self.layout = Layout(self)
        if style:
            self.style = style.copy()
        else:
            self.style = CSS()

    def __repr__(self):
        return "<%s:%s>" % (self.__class__.__name__, id(self))

    @property
    def id(self):
        '''
        The node identifier. This id can be used to target CSS directives

        :rtype: ``str``
        '''
        return self._id

    @property
    def style(self):
        '''
        The style object for this widget.

        :return: The style for this widget
        :rtype: :class:`colosseum.CSSNode`
        '''
        return self._style

    @style.setter
    def style(self, value):
        self._style = value.bind(self)

    @property
    def parent(self):
        '''
        The parent of this node.

        :rtype: :class:`toga.Widget`
        '''
        return self._parent

    @property
    def children(self):
        '''
        The children of this node.

        This *always* returns a list, even if the node is a leaf
        and cannot have children.

        :rtype: ``list``
        :return: A list of the children for this widget
        '''
        if self._children is None:
            return []
        else:
            return self._children

    def add(self, child):
        '''
        Add a widget as a child of this one.

        Raises an :class:`ValueError` if this widget is a leaf, and cannot
        have children.

        :param child: The child to add to the widget
        :type  child: :class:`toga.Widget`
        '''
        if self._children is None:
            raise ValueError('Widget cannot have children')

        self._children.append(child)

        child.app = self.app
        child._parent = self

        if self.parent:
            self.parent.layout.dirty = True

        self._add_child(child)

    @property
    def app(self):
        '''
        The App to which this widget belongs.

        :rtype: :class:`toga.App`
        '''
        return self._app

    @app.setter
    def app(self, app):
        '''
        Set the app to which this widget belongs

        :param app: The Application host
        :type  app: :class:`toga.App`
        '''
        if self._app is not None:
            if self._app != app:
                raise ValueError("Widget %s is already associated with an App" % self)
        elif app is not None:
            self._app = app
            self._set_app(app)
            if self._children is not None:
                for child in self._children:
                    child.app = app

    @property
    def window(self):
        '''
        The Window to which this widget belongs.

        :rtype: :class:`toga.Window`
        '''
        return self._window

    @window.setter
    def window(self, window):
        '''
        Set the Window to which this widget belongs.

        :param window: The new window
        :type  window: :class:`toga.Window`
        '''
        self._window = window
        self._set_window(window)
        if self._children is not None:
            for child in self._children:
                child.window = window

    @property
    def _container(self):
        '''
        The display container to which this widget belongs.
        '''
        return self.__container

    @_container.setter
    def _container(self, container):
        self.__container = container
        self._set_container(container)
        if self._children is not None:
            for child in self._children:
                child._container = container

    def _create(self):
        self.create()
        self._configure(**self._config)

    def _initialize(self, **initial):
        pass

    def _update_layout(self, **style):
        """Force a layout update on the widget.

        The update request can be accompanied by additional style information
        (probably min_width, min_height, width or height) to control the
        layout.
        """
        if self._layout_in_progress:
            return
        self._layout_in_progress = True

        if style:
            self.style.set(**style)

        # Recompute layout for this widget
        self.style.apply()

        # Update the layout parameters for all children.
        # This will also perform a leaf-first update of
        # the constraint on each widget.
        self._update_child_layout()

        # Set the constraints the widget to adhere to the new style.
        self._apply_layout()
        self._layout_in_progress = False

    def _update_child_layout(self):
        # print("UPDATE CHILD LAYOUT - widget")
        if self._children is not None:
            for child in self.children:
                # if child.is_container:
                child._update_layout()

    def set_font(self, font):
        """
        Set a font on this widget.

        :param font: The new font
        :type  font: :class:`toga.Font`
        """
        self._set_font(font)



from .base import Widget


class Box(Widget):
    '''
    This is a Widget that contains other widgets, but has no rendering or
    interaction of its own.

    :param id:          An identifier for this widget.
    :type  id:          ``str``

    :param style:       an optional style object. If no style is provided then a
                        new one will be created for the widget.
    :type style:        :class:`colosseum.CSSNode`

    :param children:    An optional list of child Widgets that will be in this
                        box.
    :type children:     ``list``
    '''
    def __init__(self, id=None, style=None, children=None):
        super().__init__(id=id, style=style, children=children)
        self._children = []

    def _configure(self, children):
        if children:
            for child in children:
                self.add(child)



class Font:
    """
    A `Font` is a font family (e.g. "Helvetica") and a size (e.g. 15) that can
    be applied to widgets.
    """
    def __init__(self, family, size):
        self._family = family
        self._size = size
        self.create()

    def create(self):
        '''
        Create font
        '''
        raise NotImplementedError(
            'Platform implementation must define create()')

    @property
    def family(self):
        '''
        Font family, e.g. Helvetica
        
        :rtype: ``str``
        '''
        return self._family

    @property
    def size(self):
        '''
        Font size
        
        :rtype: ``int``
        '''
        return self._size



from builtins import id as identifier


class App(object):
    '''
    The App is the top level of any GUI program. It is the manager of all
    the other bits of the GUI app: the main window and events that window
    generates like user input.

    When you create an App you need to provide it a name, an id for uniqueness
    (by convention, the identifier is a "reversed domain name".) and an
    optional startup function which should run once the App has initialised.
    The startup function typically constructs some initial user interface.

    Once the app is created you should invoke the main_loop() method, which
    will hand over execution of your program to Toga to make the App interface
    do its thing.

    Here is the absolute minimum App::

        app = toga.App('Empty App', 'org.pybee.empty')
        app.main_loop()
    '''
    _MAIN_WINDOW_CLASS = None
    app = None

    def __init__(self, name, app_id, icon=None,
                 id=None, startup=None, document_types=None):
        '''
        Instantiate a new application
        
        :param name: The name of the application
        :type  name: ``str``
        
        :param app_id: The unique application identifier,
            the reversed domain name, e.g. 'org.pybee.me'
        :type  app_id: ``str``
        
        :param icon: Icon for the application
        :type  icon: ``str``
        
        :param id: The DOM identifier for the app (optional)
        :type  id: ``str``
        
        :param startup: The callback method before starting the app, typically
            to add the components
        :type  startup: ``callable`` that expects a single argument of :class:`toga.App`
        
        :param document_types: Document types
        :type  document_types: ``list`` of ``str``
        '''
        App.app = self

        if self._MAIN_WINDOW_CLASS is None:
            raise NotImplementedError('App class must define _MAIN_WINDOW_CLASS')

        self.name = name
        self._app_id = app_id
        self._id = id if id else identifier(self)

        self.icon = icon

        self.document_types = document_types
        self._documents = []

        self._startup_method = startup

    @property
    def app_id(self):
        '''
        The identifier for the app.

        This is the reversed domain name, often used for targetting resources, etc.
        
        :rtype: ``str``
        '''
        return self._id

    @property
    def id(self):
        '''
        The DOM identifier for the app.

        This id can be used to target CSS directives
        
        :rtype: ``str``
        '''
        return self._id

    @property
    def documents(self):
        '''
        Return the list of documents associated with this app.
        
        :rtype: ``list`` of ``str``
        '''
        return self._documents

    def add_document(self, doc):
        '''
        Add a new document to this app.
        
        :param doc: The document to add
        '''
        doc.app = self
        self._documents.append(doc)

    def open_document(self, fileURL):
        '''
        Add a new document to this app.
        
        :param fileURL: The URL/path to the file to add as a document
        :type  fileURL: ``str``
        '''
        raise NotImplementedError('Application class must define open_document()')

    def startup(self):
        '''
        Create and show the main window for the application
        '''
        self.main_window = self._MAIN_WINDOW_CLASS(self.name)
        self.main_window.app = self

        if self._startup_method:
            self.main_window.content = self._startup_method(self)

        self.main_window.show()

    def main_loop(self):
        '''
        Invoke the application to handle user input.

        This method typically only returns once the application is exiting.
        '''
        raise NotImplementedError('Application class must define main_loop()')



from .base import Widget


class NavigationView(Widget):
    def __init__(self, title, content, on_action=None, style=None):
        super().__init__(title=title, content=content, on_action=on_action, style=style)

    def _configure(self, title, content, on_action):
        self.title = title
        self.content = [content]
        self.on_action = on_action

    def push(self, content):
        self.content.push(content)
        self._push(content)

    def _pop(self):
        self.content.pop()
        self._pop()



import os


class Icon:
    app_icon = None

    def __init__(self, path, system=False):
        if os.path.splitext(path)[1] in ('.png', '.icns', '.bmp'):
            self.path = path
        else:
            self.path = path + self.EXTENSION
        self.system = system


        if self.system:
            toga_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

            filename = os.path.join(toga_dir, 'resources', self.path)
        else:
            filename = self.path

        self.create(filename)

    def create(self, filename):
        raise NotImplementedError('Icon must define create()')

    @classmethod
    def load(cls, path_or_icon, default=None):
        if path_or_icon:
            if isinstance(path_or_icon, Icon):
                obj = path_or_icon
            else:
                obj = cls(path_or_icon)
        elif default:
            obj = default
        return obj



from .base import Widget


class ProgressBar(Widget):
    '''
    Progressbar widget
    '''
    def __init__(self, id=None, style=None, max=None, value=None):
        '''
        Instantiate a new instance of the progress bar widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param max: The maximum value
        :type  max: ``int``

        :param value: The initial value
        :type  value: ``int``
        '''
        super().__init__(id=id, style=style, max=max, value=value)

    def _configure(self, max, value):
        self.value = value
        self.max = max
        self.running = False
        self.rehint()

    @property
    def value(self):
        '''
        The progress value
        
        :rtype: ``int``
        '''
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        self._running = self._value is not None

    @property
    def max(self):
        '''
        The maximum value

        :rtype: ``int``
        '''
        return self._max

    @max.setter
    def max(self, max):
        self._max = max
        self._set_max(max)



from .base import Widget


class MultilineTextInput(Widget):
    '''
    A multi-line text input widget
    '''
    def __init__(self, id=None, style=None, initial=None):
        '''
        Instantiate a new instance of the multi-line text input widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param initial:     The initial value
        :type  initial:     ``str``
        '''
        super().__init__(id=id, style=style, initial=initial)



from .base import Widget


class OptionContainer(Widget):
    _CONTAINER_CLASS = None

    def __init__(self, id=None, style=None, content=None):
        '''
        Instantiate a new instance of the option container widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`
        
        :param content:     List of components to choose from.
        :type  content:     ``list`` of ``tuple`` (``str``, :class:`toga.Widget`)
        '''
        
        super().__init__(id=id, style=style, content=content)
        self._containers = []

    def _configure(self, content):
        if content:
            for label, widget in content:
                self.add(label, widget)

    def add(self, label, widget):
        '''
        Add a widget to the option container
        
        :param label: The label for the option
        :type  label: ``str``
        
        :param widget: The widget to add
        :type  widget: :class:`toga.Widget`
        '''
        if widget._impl is None:
            container = self._CONTAINER_CLASS()
            container.content = widget
        else:
            container = widget

        self._containers.append((label, container, widget))

        widget.window = self.window
        widget.app = self.app

        self._add_content(label, container, widget)

    def _update_child_layout(self):
        for label, container, widget in self._containers:
            container._update_layout()



from .base import Widget


class Tree(Widget):
    '''
    Tree widget
    '''
    def __init__(self, headings, id=None, style=None):
        '''
        Instantiate a new instance of the tree widget

        :param headings: The list of headings for the tree
        :type  headings: ``list`` of ``str``

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`
        '''
        super().__init__(id=id, style=style)
        self.headings = headings

    def _configure(self):
        pass



from .base import Widget


class ImageView(Widget):
    '''
    An image viewer
    '''
    def __init__(self, id=None, style=None, image=None):
        '''
        Instantiate a new image viewer widget
        
        :param id:          An identifier for this widget.
        :type  id:          ``str``
    
        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`
    
        :param image:    Image to display
        '''
        super().__init__(id=id, style=style, image=image)

    def _configure(self, image):
        self.image = image

    # @property
    # def alignment(self):
    #     return self._alignment

    # @alignment.setter
    # def alignment(self, value):
    #     self._alignment = value
    #     self._impl.setAlignment_(NSTextAlignment(self._alignment))

    # @property
    # def scaling(self):
    #     return self._scaling

    # @scaling.setter
    # def scaling(self, value):
    #     self._scaling = value
    #     self._impl.setAlignment_(NSTextAlignment(self._scaling))
